
/*
 * libgnet - game net library
 * Copyright (C) 1997 Indrek Mandre 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Library General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>

#include "gnet.h"
#include "gnet-local.h"

void GN_close_conn ( GN_conn *ppp )
{

 close ( ppp->socket );
 if ( ppp->udpsocket != -1 ) close ( ppp->udpsocket );
 if ( ppp->sendtoudp != NULL ) free ( ppp->sendtoudp );

}

GN_conn * GN_connect ( int sock_type, char *host, int port, int sock_arg )
{
 GN_conn *con;
 struct GN_connection bb;

 switch ( sock_type ) {
 	case NET_UNIX:
 	{
 		struct sockaddr_un sunix;
 		int s;

		if ( sock_arg != 0 ) {
			sprintf ( GN_error, "Invalid sock_arg %d to UNIX socket\n", sock_arg );
			return NULL;
		}

 		memset ( &sunix, 0, sizeof ( struct sockaddr_un ) );

 		sunix.sun_family = AF_UNIX;
 		if ( GN_unix_socket_name == NULL ) 
 			sprintf ( sunix.sun_path, "/tmp/GN%d", port );
 		else strncpy ( sunix.sun_path, GN_unix_socket_name, UNIX_PATH_MAX );

		if ( ( s = socket ( AF_UNIX, SOCK_STREAM, 0 ) ) == -1 ) {
			sprintf ( GN_error, "Error creating AF_UNIX type socket, errno=%s\n", strerror ( errno ) );
			return NULL;
		}

		if ( connect ( s, (struct sockaddr *)&sunix, sizeof ( struct sockaddr_un ) ) == -1 ) {
			sprintf ( GN_error, "Couldn't connect to AF_UNIX type socket, path %s, errno=%s\n", sunix.sun_path, strerror ( errno ) );
			close (s);
			return NULL;
		}

		if ( recv ( s, &bb, sizeof ( struct GN_connection ), 0 ) == -1 ) {
			sprintf ( GN_error, "Strange error with recv, errno=%s\n", strerror (errno) );
			close (s);
			return NULL;
		}

		bb.version = GNET_VERSION;
		bb.udpport = -1;

		send ( s, &bb, sizeof ( struct GN_connection ), 0 );

		con = malloc ( sizeof ( GN_conn ) );
		con->socket = s;
		con->udpport = 0;
		con->sendtoudp = NULL;

		return con;

 		break;
	}

	case NET_INTERNET:
	{
		struct sockaddr_in sinet;
		struct hostent *hp;
		int u = -1;
		int s, s2 = -1;

		memset ( &sinet, 0, sizeof ( struct sockaddr_in ) );

		sinet.sin_family = AF_INET;
		sinet.sin_port = htons ( port );

		hp = gethostbyname ( host );

		if ( hp == NULL ) {
			sprintf ( GN_error, "Couldn't find host %s, errno=%s\n", host, strerror ( errno ) );
			return NULL;
		}

		sinet.sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;

		if ( ( s = socket ( AF_INET, sock_arg&CON_SEQPACKET?SOCK_SEQPACKET:SOCK_STREAM, 0 ) ) == -1 ) {
			sprintf ( GN_error, "Error creating AF_INET type socket, errno=%s\n", strerror ( errno ) );
			return NULL;
		}

		if ( connect ( s, (struct sockaddr *) &sinet, sizeof ( struct sockaddr_in ) ) == -1 ) {
			sprintf ( GN_error, "Error connecting socket to %s %d, errno=%s\n", host, port, strerror ( errno ) );
			close ( s );
			return NULL;
		}

		if ( recv ( s, &bb, sizeof ( struct GN_connection ), 0 ) == -1 ) {
			sprintf ( GN_error, "Strange error with recv, errno=%s\n", strerror (errno) );
			close (s);
			return NULL;
		}

		if ( bb.udpport != -1 ) {
			struct sockaddr_in udp_in;
			int len;

			u = bb.udpport;

			if ( ( s2 = socket ( AF_INET, SOCK_DGRAM, 0 ) ) == -1 ) {
				sprintf ( GN_error, "Couldn't create UDP type socket, errno=%s\n", strerror ( errno ) );
				close (s);
				return NULL;
			}

			udp_in.sin_family = AF_INET;
			udp_in.sin_addr.s_addr = INADDR_ANY;
			udp_in.sin_port = 0;

			len = sizeof ( struct sockaddr_in );

			if ( bind ( s2, (struct sockaddr *)&udp_in, len ) == -1 ) {
				sprintf ( GN_error, "Couldn't bind udp socket, errno=%s\n", strerror ( errno ) );
				close (s2);
				close (s);
				return NULL;
			}

			getsockname ( s2, (struct sockaddr *) &udp_in, &len );

			bb.udpport = htons ( udp_in.sin_port );
		}

		bb.version = GNET_VERSION;

		send ( s, &bb, sizeof ( struct GN_connection ), 0 );

		con = malloc ( sizeof ( GN_conn ) );
		con->socket = s;
		con->udpport = bb.udpport;
		con->sendtoudp = NULL;
		con->udpsocket = s2;

		if ( u != -1 ) {
			con->sendtoudp = malloc ( sizeof ( struct sockaddr_in ) );
			memset ( con->sendtoudp, 0, sizeof ( struct sockaddr_in ) );
			((struct sockaddr_in *)con->sendtoudp)->sin_family = AF_INET;
			((struct sockaddr_in *)con->sendtoudp)->sin_port = htons ( u );
			((struct sockaddr_in *)con->sendtoudp)->sin_addr.s_addr = ((struct in_addr *)(hp->h_addr))->s_addr;
		}

		return con;

		break;
	}
	default:
		sprintf ( GN_error, "Unknown socket type %d\n",  sock_type );
		return NULL;
 }

 con = malloc ( sizeof ( GN_conn ) );
 return con;
}

